package grm;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import org.jacorb.transaction.Sleeper;
import org.omg.CORBA.ORB;

import subtypes.ApplicationType;
import subtypes.AsctRequestId;
import types.BspInfo;
import types.CommonExecutionSpecs;
import types.DistinctExecutionSpecs;
import interfaces.ExecutionManagerPOA;
import interfaces.Grm;
import interfaces.Lrm;
import interfaces.LrmHelper;

/** Flag to display debugging info. */
interface ExecManDebug {
    public static final boolean Debug = false;
}

public class ExecutionManagerImpl extends ExecutionManagerPOA {

    
    class AppInfo {
        /** AppNodeId x string[] {NodeLRM, AppId} */
        Map<Integer, String[]> nodeAppIds;
        CommonExecutionSpecs commonSpecs;   
        DistinctExecutionSpecs[] distinctSpecs;        
        boolean forceDifferentMachines;
        BspRestartCoordinator bspCoord;
     
        int currentRestart = -1;
        boolean isRestarting = false;
        boolean isAborting   = false;
        
        AtomicInteger nRestarts  = new AtomicInteger(0);
        AtomicInteger nResponses = new AtomicInteger(0);
        //final Lock nodeAppIdsLock = new ReentrantLock();
        Lock isActive;
        Condition nextRestartCond;        
        Lock waitingResponses;
        Condition finishedResponsesCond;
    }
    
    ConcurrentHashMap<String, AppInfo> currentExecutions;
    Grm grm; // Reference to the GRM corba object
    ORB orb; // Reference to the ORB
    BspFinishedExecutionManager bspManager; 
    StandardFinishedExecutionManager standardManager;
    
    public ExecutionManagerImpl(ORB orb, Grm grm) {
        this.orb = orb;
        this.grm = grm;
        this.currentExecutions = new ConcurrentHashMap<String, AppInfo>();
        this.bspManager        = new BspFinishedExecutionManager(currentExecutions, orb, grm);
        this.standardManager   = new StandardFinishedExecutionManager(currentExecutions, orb, grm);
    }

    /** ---------------------------------------------------------------------------
     *  ExecutionManager Methods
     *  ---------------------------------------------------------------------------
     * */

    public void reportExecutionScheduled(CommonExecutionSpecs commonSpecs,
            DistinctExecutionSpecs[] distinctSpecs, boolean forceDifferentMachines) {
        
        if (ExecManDebug.Debug)
            System.err.println(">>>>> ExecutionManagerImpl.reportExecutionScheduled--> appMainRequestId:" + distinctSpecs[0].asctRequestId.appMainRequestId);        
                
        String appMainRequestId = distinctSpecs[0].asctRequestId.appMainRequestId;        
        
        AppInfo appInfo = new AppInfo();
        if (currentExecutions.putIfAbsent(appMainRequestId, appInfo) == null) {                        
            appInfo.commonSpecs   = commonSpecs;
            appInfo.distinctSpecs = distinctSpecs;
            appInfo.forceDifferentMachines = forceDifferentMachines;
            appInfo.nodeAppIds    = new HashMap<Integer, String[]>(distinctSpecs.length);

            if (commonSpecs.appType == ApplicationType.bsp) {
                appInfo.bspCoord = new BspRestartCoordinator(orb, grm, distinctSpecs.length);      
                appInfo.isActive = new ReentrantLock();
                appInfo.nextRestartCond = appInfo.isActive.newCondition();
                appInfo.waitingResponses = new ReentrantLock();
                appInfo.finishedResponsesCond = appInfo.waitingResponses.newCondition();                
            }
        }

    }

    //--------------------------------------------------------------------------------------
    
    public void reportExecutionStarted(String lrmIor, String appId,
            int restartId, AsctRequestId asctRequestId) {
                
        AppInfo appInfo = currentExecutions.get(asctRequestId.appMainRequestId);
        if (appInfo == null) {
            /** TODO: We should treat this error better. It would be a good idea to kill the process in the LRM. */
            System.err.println("ExecutionManagerImpl.reportExecutionStarted --> ERROR: Couldn't find execution info.");
            System.err.println("appId:" + appId + " asctRequestId:" + asctRequestId.appMainRequestId + "|" + asctRequestId.appNodeRequestId);
        }

        int nodeId = Integer.parseInt(asctRequestId.appNodeRequestId);
        appInfo.nodeAppIds.put(nodeId, new String[] {lrmIor, appId});
        int nRestarts = appInfo.nRestarts.incrementAndGet();

        if (ExecManDebug.Debug)
            System.err.println(">>>>> ExecutionManagerImpl.reportExecutionStarted-->asctRequestId:" + asctRequestId.appMainRequestId + "|" + asctRequestId.appNodeRequestId +  ", nRestarts=" + nRestarts);       
        
        // Waits for all responses from appInfo 
        while (appInfo.isRestarting == true && appInfo.nResponses.get() < appInfo.distinctSpecs.length) {
            try {Thread.sleep(100);} 
            catch (InterruptedException e) {}
        }
            
        if (nRestarts == appInfo.distinctSpecs.length) {
            System.out.println("ExecutionManagerImpl.reportExecutionStarted--> Execution Started [" + asctRequestId.appMainRequestId + "," + restartId + "]");                
            appInfo.currentRestart++;
            if (appInfo.isActive != null) {
                appInfo.isActive.lock();
                try{
                    if (ExecManDebug.Debug)
                        System.out.println("Finished Starting");
                    appInfo.isRestarting = false;
                    appInfo.nextRestartCond.signalAll();
                }
                finally{ appInfo.isActive.unlock(); }
            }
        }
 
    }

    //--------------------------------------------------------------------------------------
    
    public int reportExecutionFinished(int restartId, AsctRequestId asctRequestId, int status) {
                
        if (ExecManDebug.Debug)
            System.err.println(">>>>> ExecutionManagerImpl.reportExecutionFinished-->asctRequestId: " + asctRequestId.appMainRequestId + "|" + asctRequestId.appNodeRequestId + " exitCode:" + status + " restart: {" + restartId + "}");

        /** Get the appInfo for the finished process */
        AppInfo appInfo = currentExecutions.get(asctRequestId.appMainRequestId);
        if (appInfo == null) {
            System.err.println("ExecutionManagerImpl.reportExecutionStarted --> ERROR: No existing execution." + " asctRequestId:" + asctRequestId.appMainRequestId + "|" + asctRequestId.appNodeRequestId);
            return -1;
        }
        
        if (appInfo.commonSpecs.appType == ApplicationType.bsp)
            bspManager.treatFinishedExecution(appInfo, asctRequestId, restartId, status);
        else
            standardManager.treatFinishedExecution(appInfo, asctRequestId, restartId, status);
        
        return 0;
    }

    //--------------------------------------------------------------------------------------
    
    public BspInfo registerBspNode(String asctIor, AsctRequestId asctRequestId,
            String bspProxyIor) {
        
        /** Get the appInfo for the finished process */
        AppInfo appInfo = currentExecutions.get(asctRequestId.appMainRequestId);
        if (appInfo == null) {
            System.err.println("ExecutionManagerImpl.registerBspNode --> ERROR: No existing execution." + " asctRequestId:" + asctRequestId.appMainRequestId + "|" + asctRequestId.appNodeRequestId);
            return new BspInfo();
        }
        
        return appInfo.bspCoord.registerBspNode(asctRequestId, bspProxyIor);
    }
    
    //--------------------------------------------------------------------------------------
}
